from dataclasses import dataclass
import pandas as pd
import numpy as np
import os
import time
import warnings
import webbrowser

from .utils import parse_date_parameter
from .scrape import CHARTFREQUENCIES
from .utils import findColumnTitles, fixCommas, is_leap_year
from .fetch import FIRST_OBSERVATION

BCCR_FOLDER = os.path.dirname(os.path.abspath(__file__))
DATA_FOLDER = os.path.join(BCCR_FOLDER, 'data')
PICKLE_FILE = os.path.join(DATA_FOLDER, 'cuadros.pkl')

FRASE_AYUDA = """
CLASE PaginaWeb

Esta clase permite buscar y descargar cuadros de indicadores del sitio web del Banco Central de Costa Rica.
Suponiendo que el objeto de clase PaginaWeb se llama "consulta":
    * para buscar cuadros, utilice 
        consulta.buscar()
    * para descargar datos de cuadros 4, 7 y 231 (por ejemplo), hay varias formas de hacerlo 
        consulta(4, 7, 231)   # pasando los códigos directamente
        consulta([4, 7, 231]) # pasando los códigos en una lista
        consulta({'4':'indicA', '7':'indicB', '231':'indicC'} # pasando los códigos en un diccionario        
"""

@dataclass
class PaginaWeb:
    cuadros: pd.DataFrame

    def api(self, Cuadro, *, FechaInicio=None, FechaFinal=None, excel=True, abrir=False):
        """
        Construye un URL para obtener datos del sitio de indicadores económicos del BCCR

        Parameters
        ----------
        Cuadro:  int
              Número que identifica el cuadro que desea descargarse del sitio web del BCCR
        FechaInicio : str or int, optional
            fecha de la primera observación a descargar. Ver nota 1 de la función `datos` para detalles de formato.
        FechaFinal : str or int, optional
            fecha de la última observación a descargar. Ver nota 1 de la función `datos` para detalles de formato.
        excel: bool (opcional, predeterminado = True)
            Si `True`, obtiene la versión Excel de los datos, `False` obtiene la versión HTML
        abrir: bool (opcional, predeterminado = True)
            Si `True`, abre la página web indicada por el URL en el navegador de internet predeterminado
        Returns
        -------
            El URL para descargar los datos del cuadro indicado: str

        Examples
        --------
            1. Obtener el URL  del IPC (Cuadro 9)

            >>> from bccr import PW
            >>> PW.api(9)

            'https://gee.bccr.fi.cr/indicadoreseconomicos/Cuadros/frmVerCatCuadro.aspx?CodCuadro=9&Exportar=True&Excel=True'

            2. URL para datos del M1 desde enero de 2010

            >>> PW.api(125, FechaInicio=2010)

            'https://gee.bccr.fi.cr/indicadoreseconomicos/Cuadros/frmVerCatCuadro.aspx?CodCuadro=125&FecInicial=2010/01/01&Exportar=True&Excel=True'

            3. URL para datos del IPC no transable (Cuadro 289) entre 2010 y 2015

            >>> PW.api(289, FechaInicio=2010, FechaFinal=2015)

            'https://gee.bccr.fi.cr/indicadoreseconomicos/Cuadros/frmVerCatCuadro.aspx?CodCuadro=289&FecInicial=2010/01/01&FecFinal=2015/12/31&Exportar=True&Excel=True'

            4. URL para el M1 entre 2010 y 2015, pero en formato HTML (en vez del formato Excel predeterminado)

            >>> PW.api(125, FechaInicio=2010, FechaFinal=2015, excel=False)

            'https://gee.bccr.fi.cr/indicadoreseconomicos/Cuadros/frmVerCatCuadro.aspx?CodCuadro=125&FecInicial=2010/01/01&FecFinal=2015/12/31'

            5. Igual que el anterior, pero usando fechas predeterminadas. La opción `abrir=True` muestra los datos en su navegador de internet predeterminado

            >>> PW.api(125, excel=False, abrir=True)  # opens link in browser

            'https://gee.bccr.fi.cr/indicadoreseconomicos/Cuadros/frmVerCatCuadro.aspx?CodCuadro=125'
        """
        bccr_web = "https://gee.bccr.fi.cr/indicadoreseconomicos/"

        bccr_web += "Cuadros/frmVerCatCuadro.aspx?"
        params = dict(CodCuadro=Cuadro)
        if FechaInicio:
            params['FecInicial'] = parse_date_parameter(FechaInicio, inicio=True, año_primero=True)
        if FechaFinal:
            params['FecFinal'] = parse_date_parameter(FechaFinal, inicio=False, año_primero=True)
        if excel:
            params['Exportar'] = True
            params['Excel'] = True

        bccr_web += "&".join(f"{k}={v}" for k,v in params.items())

        if abrir:
            webbrowser.open(bccr_web)

        return bccr_web

    def web(self, Cuadro):
        """
        Abre el cuadro especificado en el sitio del BCCR, usando el navegador de Internet predeterminado.

        Parameters
        ----------
        Cuadro:  int or str
             Número de cuadro que se desea verificar

        Returns
        -------
            None

        Examples
        --------
        Para abrir el cuadro 125

        >>> self.web(125)
        """

        self.api(Cuadro, excel=False, abrir=True)
        return

    def __downloadChart__(self, chart, first=None, last=None, quiet=False):
        """
            Downloads data from BCCR website.

            This function converts a BCCR html file into a pandas dataframe. Other
            functions take this dataframe and "clean it" to return a tidy dataframe.

        Parameters
        ----------
        chart   : A number identifying the BCCR's data table (integer).
        first   : The first year to download (integer, default=None).
        last    : The last year to download (integer, default=None)
        quiet   : Print download info if False, nothing if True

        Returns
        -------
            data    : A pandas DataFrame with data in the same format as in BCCR website. First column is used to set the
                    (row) index and column headers are use to set the columns property
            title   : Chart title (string)
            subtitle : Chart subtitles (string)

        Examples
        --------
            1. Download data on money supply (M1) for 2011 to 2015

            >>> self.__downloadChart__(125, 2011, 2015)

            2. Same as before, but without printing download info(name of series, url, retrieved date-time)

            >>> self.__downloadChart__(138, 2011, 2015, quiet=True)
        """
        chart_url = self.api(chart, FechaInicio=first, FechaFinal=last,excel=True, abrir=False)
        rawdata = pd.read_html(chart_url, thousands="")[0]
        title, subtitle, subts2 = rawdata.iloc[:3, 0]
        if pd.notnull(subts2):
            subtitle += ' --- %s' % subts2

        rawdata.set_index(0, inplace=True)
        h = findColumnTitles(rawdata)
        rawdata.columns = rawdata.iloc[h]
        rawdata = rawdata.iloc[h + 1:].applymap(fixCommas)

        if not quiet:
            info = 'Descargando el cuadro %s:' % chart
            info += ('\n\t' + title) if title else ''
            info += ('\n\t' + subtitle) if subtitle else ''
            info += '\n\tDescargado el %s desde:' % time.strftime("%c")
            info += '\n\t' + chart_url + '\n'
            print(info)

        rawdata._metadata = {'title': title, 'subtitle': subtitle}
        return rawdata

    def __parse__(self, chart, first=None, last=None, freq=None, func=None, quiet=True):

        # TODO: Add chartFormat='DayIndicator' to the possible options!!! Example chart=572

        chartFormat = self.cuadros.loc[chart, 'chartFormat']

        data = self.__downloadChart__(chart, first, last, quiet)
        title = data._metadata['title']

        ''' CLEAN DATA'''
        if chartFormat == 'MonthYear' and 'total' in str(data.index[0]).lower():
            data.drop(data.index[0], inplace=True)
        elif chartFormat == 'DayYear':
            nonleap = np.array([~is_leap_year(int(y)) for y in data.columns])
            data.iloc[59, nonleap] = np.inf  # row 59 = Feb 29 (counting from zero-base)

        ''' GET FIRST OBSERVATION '''
        try:
            t0 = FIRST_OBSERVATION[chartFormat](data)
        except:
            raise Exception(NotImplemented)

        ''' GET DATA IN TIME v. SERIES FORMAT '''
        if chartFormat in ['IndicatorYear', 'IndicatorQuarter', 'IndicatorMonth', 'MonthYear', 'DayYear']:
            data = data.transpose()

        ''' STACK DATA '''
        if chartFormat in ['YearMonth', 'MonthYear', 'DayYear']:
            data = data.stack(dropna=False)

        ''' REMOVE 29 FEB IN NON-LEAP YEARS '''
        if chartFormat in ['DayYear']:
            data = data[data != np.inf]

        ''' REMOVE 29 REB IN NON-LEAP YEARS'''
        if chartFormat in ['DayIndicator']:
            not_a_leap_year = ~is_leap_year(np.array([int(x.split()[-1]) for x in data.index]))
            feb29 = np.array(['29 Feb' in x for x in data.index])
            wrong_day = feb29 & not_a_leap_year
            data = data.iloc[~wrong_day]

        ''' ADD TIME INDEX '''
        data.index = pd.period_range(start=t0, periods=data.shape[0], freq=CHARTFREQUENCIES[chartFormat][0])

        ''' DROP MISSING VALUES '''
        if isinstance(data, pd.Series):
            data.name = str(chart)
            data.dropna(inplace=True)
        elif isinstance(data, pd.DataFrame):
            data.columns = [str(x) for x in data.columns]
            data = data.dropna(axis=0, how='all').dropna(axis=1, how='all')
        else:
            raise ValueError('Unexpected data type: %s' % type(data))

        ''' RESAMPLE DATA '''
        if freq:
            func = func if func else np.mean
            data = data.resample(freq).apply(func)

        return data

    def soporte(self, cuadro):
        """
        Indica si un `cuadro` específico puede descargarse con PaginaWeb

        Parameters
        ----------
        cuadro: int o str
            Número de cuadro que se desea verificar

        Returns
        -------
            bool

        """
        return int(cuadro) in self.cuadros.index


    def __nombre__(self, cuadro):
        """
        Nombre del indicador con ese código en el catálogo. Si no existe, entonces el código mismo
        Parameters
        ----------
        codigo int or str, código del indicador

        Returns
        -------
        str
        """
        return self.cuadros.loc[cuadro, 'title'] if self.soporte(cuadro) else str(cuadro)


    def datos(self, *Cuadros, FechaInicio=None, FechaFinal=None, func=np.sum, freq=None, info=False, **indicadores):
        """ Descargar datos del sitio de Indicadores Económicos del BCCR

        Construye una consulta por método GET a partir de los parámetros proporcionados, para cada uno de los
        Indicadores solicitados. Descarga los datos, los transforma en una tabla de datos de Pandas.
        Si hay indicadores de distintas frecuecias, los transforma a la misma frecuencia según el método indicado.

        Parameters
        ----------
        Cuadros: sucesión de enteros o strings
                Los códigos numéricos (como int o str) de los cuadros que se desean descargar (ver parámetro
                `indicadores`). En el resultado, el indicador se identifica con el nombre que tiene en el catálogo de
                cuentas; en caso de querer indicar un nombre distinto, solicitar el cuadro usando el parámetro
                `indicadores`
        FechaInicio : str or int, optional
            fecha de la primera observación a descargar. Ver nota 6 abajo para más detalles.
        FechaFinal : str or int, optional
            fecha de la última observación a descargar. Ver nota 6 abajo para más detalles.
        func
        freq : str, optional
            frecuencia a la que se quiere convertir los datos. Si este parámetro no se indica y se requieren series de
            distintas frecuencias, el resultado será un Data.Frame con la menor periodicidad de las series solicitadas.
        info:  bool, optional (default=False)
            Si `True`, se imprime metadata del cuadro descargado, así como un hipervículo para consultar los datos en el
            sitio de Internet del BCCR
        indicadores: pares de nombre=codigo, de los cuadros requeridos
             Este es el método preferible para indicar las series requeridas (en vez del parámetro `Códigos`). Las series
             se enumeran como pares de nombre=codigo, donde nombre es el nombre que tendrá el indicador en el DataFrame
             resultante, y codigo es un número entero que identifica el cuadro en la página de indicadores económicos del BCCR.
        Returns
        -------
        pd.DataFrame
            una tabla con los datos solicitados, filas indexadas por tiempo, cada columna es un indicador.

        Notes
        -----
        1. Esta función construye un URL a partir de los parámetros proporcionados, para cada uno de los Indicadores solicitados.
        Descarga los datos, los transforma en una tabla de datos de Pandas.

        2. Si hay indicadores de distintas frecuecias, los transforma a la misma frecuencia según el método indicado.

        3. A excepción de `*Cuadros`, todos los parámetros deben usar palabra clave, aunque es preferible usar el parámetro
        `indicadores` en vez de `*Cuadros`. (Ver los ejemplos)

        4. Hay varias maneras de indicar los Indicadores: (a) simplemente enumerándolos, (b) en una lista o tupla, y
        (c) en un diccionario (método obsoleto, es mejor usar el parámetro `indicadores`). En caso de tratarse de
        diccionario, los códigos se indican como las llaves del diccionario, mientras que los valores del diccionario
        se usan para renombrar las columnas resultantes.

        5. Las instancias de la clase PaginaWeb son ejecutables: si se llaman como una función, simplemente ejecutan la
        función datos()

        6. El formato de fechas es muy flexible. Por ejemplo, todas estas expresiones son válidas:

        - Para indicar el año 2015: se puede emplear tanto un `int` como un `str`

        >>>   FechaInicio = 2015 # se interpreta como  1 de enero de 2015
        >>>   FechaFinal = "2015"  # se interpreta como 31 de diciembre de 2015

        - Para indicar el mes marzo de 2017: cualquiera de estas expresiones es válida

         >>> "2017-03"
         >>> "2017/03"
         >>> "2017m3"
         >>> "03/2017"
         >>> "03-2017"

         En `FechaInicio=` resulta en 1 de marzo de 2017, en `FechaFinal=` resulta en 31 de marzo de 2017.

        - Para indicar el 12 de agosto de 2018:

         >>> "2017/8/12"
         >>> "2017-08-12"
         >>> "12/8/2017"

        Observe que para separar los componentes de una fecha se puede usar cualquier caracter no numérico (usualmente `/` o `-`).

        Examples
        --------
        La forma preferible de solicitar los indicadores es como pares de `nombre=cuadro`, de manera que `nombre` se utilice
        como encabezado de columna en la tabla de datos resultante:

        >>> from bccr import PW
        >>> SW(M1=125, Npp=177)

        """

        msg = """
        En una futura versión, los indicadores deberán ser solicitados como pares de 'nombre=codigo', o bien 
        como un listado de enteros. Por ejemplo
            PW(M1=125, Npp=177)
        o bien
            PW(125, 177)                
        """


        for indic in Cuadros:
            if isinstance(indic, dict):
                warnings.warn(msg)
                for key, value in indic.items():
                    indicadores[value] = int(key)
            elif isinstance(indic, str) or isinstance(indic, int):
                indicadores[self.__nombre__(indic)] = int(indic)
            elif hasattr(indic, '__iter__'):
                warnings.warn(msg)
                for val in indic:
                    indicadores[self.__nombre__(val)] = int(val)
            else:
                raise('No sé cómo interpretar el indicador requerido')

        indicadores_válidos = dict()
        for nombre, codigo in indicadores.items():
            if self.soporte(codigo):
                indicadores_válidos[nombre] = codigo
            else:
                print(f"PaginaWeb aún no tiene soporte para el cuadro {codigo}, o bien no existe")

        if indicadores_válidos:
            #__parse__(self, chart, first=None, last=None, freq=None, func=None, quiet=True):
            datos = {nombre: self.__parse__(codigo, first=FechaInicio, last=FechaFinal, freq=freq, func=func, quiet=not info) for nombre, codigo in indicadores_válidos.items()}

            freqs = pd.Series({codigo: self.cuadros.loc[codigo, 'freq'] for codigo in indicadores_válidos.values()})
            freqs = freqs.astype('category').cat.set_categories(['A', '6M', 'Q', 'M', 'W', 'D'], ordered=True)

            if len(freqs)>1:  # es necesario convertir frecuencias
                freq = freq if freq else freqs.min()
                if callable(func):
                    func = {codigo: func for codigo in indicadores_válidos.values()}

                for nombre, codigo in indicadores_válidos.items():
                    if freqs[codigo] != freq:
                        datos[nombre] = datos[nombre].resample(freq).apply(func[codigo])

            results = pd.concat(datos.values(), keys=datos.keys(), axis=1)

            return results


    def __call__(self, *args, **kwargs):
        return self.datos(*args, **kwargs)

    def __buscar_frase__(self, frase):
        CAMPOS = ['title', 'subtitle']
        return pd.DataFrame([self.cuadros[campo].str.contains(frase, case=False) for campo in CAMPOS]).any()

    def buscar(self, todos=None, *, frase=None, algunos=None, frecuencia=None):
        """buscar códigos de indicadores según su descripción

        Parameters
        ----------
        frase, todos, algunos : str
            texto que debe aparecer en la descripción del indicador. Sólo una de estos tres parámetros debe utilizarse
            a la vez. 'frase' busca una coincidencia exacta, 'todos' que todas las palabras aparezcan (en cualquier
            orden), 'algunos' que al menos una de las palabras aparezca. La búsqueda no es sensible a mayúscula/minúscula.
            Si no se indica un parámetro [por ejemplo, PW.buscar('precios consumidor')], se asume que
            PW.buscar(todos='precio consumidor')
        frecuencia : str, optional. uno de ('Q', 'A', 'M', 'D')
            mostrar solo indicadores que tengan la frecuencia indicada.

        Returns
        -------
        pd.DataFrame

        Examples
        --------

        Para buscar un indicador que tenga todas las palabras "Índice", "precios", "consumidor" (en cualquier orden)

        >>> from bccr import PW
        >>> PW.buscar(todos="Índice precios consumidor")

        Para buscar un indicador que tenga la frase exacta "Índice de Precios al Consumidor"

        >>> PW.buscar(frase="Índice de Precios al Consumidor")

        Para buscar un indicador que tenga al menos una de las palabras "exportaciones" e "importaciones"

        >>> PW.buscar(algunos="exportaciones importaciones")

        Para buscar los términos "precios transables", y filtrar que muestre solo resultados con medida "Variación interanual"

        >>> PW.buscar('precios transables', Medida='Variación interanual')

        Para mostrar un breve mensaje de ayuda

        >>> PW.buscar()


        """


        CAMPOS = ['title', 'subtitle', 'freq']

        if frase:
            temp = self.__buscar_frase__(frase)
        elif todos:
            temp = pd.DataFrame([self.__buscar_frase__(palabra) for palabra in todos.split(' ')]).all()
        elif algunos:
            temp = pd.DataFrame([self.__buscar_frase__(palabra) for palabra in algunos.split(' ')]).any()
        else:
            ayuda = """ BUSCAR
            Esta función ayuda a buscar los códigos de indicadores, utilizando palabras descriptivas.
            Exactamente un parámetro de [frase, todos, algunos] debe ser proporcionado.

            Ejemplos de uso:
                buscar(frase="descripción contiene esta frase literalmente")
                buscar(todos="descripción contiene todos estos términos en cualquir orden")
                buscar(algunos="descripción contiene alguno de estos términos")
                buscar()  # muestra este mensaje de ayuda
            """
            print(ayuda)
            return


        if frecuencia:
            freq = self.cuadros['freq'] == frecuencia[0].upper()
            return self.cuadros[temp & freq][CAMPOS]
        else:
            return self.cuadros[temp][CAMPOS]

    def __str__(self):
        return FRASE_AYUDA

    def __repr__(self):
        return self.__str__()


PW = PaginaWeb(pd.read_pickle(PICKLE_FILE))


import re
import numpy as np
import pandas as pd
from datetime import date
from pandas import offsets

def findFirstElement(pattern: str, stringList):
    idx = 0

    for element in list(stringList):
        if re.match(pattern, str(element), re.IGNORECASE):
            return idx
        else:
            idx += 1

    raise ValueError('No match found!')


def fixCommas(x):
    return float(x.replace(".", "").replace(",", ".")) if isinstance(x, str) else x


def resample(series, freq, func):
    if func is None:
        func = np.mean
    return series.resample(freq).apply(func) if freq else series



def tidy(data, timeindex, freq, func, colnames):

    data.index = timeindex
    if isinstance(data, pd.Series):
        data = data.apply(fixCommas)
        data.name = colnames
        data.dropna(inplace=True)
    elif isinstance(data, pd.DataFrame):
        data.columns = colnames
        data = data.applymap(fixCommas).dropna(0, 'all').dropna(1, 'all')
        #data = removeEmptyColumns(data)

    else:
        raise NotImplemented

    return resample(data, freq, func)


def parseSeriesFreqInputs(series, func):

    # make a dictionary with series
    if isinstance(series, int):
        seriesdict = {series: ''}
    elif isinstance(series, pd.Series):
        seriesdict = dict(zip(series.index, series))
    elif isinstance(series, list) or isinstance(series, tuple):
        seriesdict = {chartNumber: '' for chartNumber in series}
    elif isinstance(series, dict):
        seriesdict = series
    else:
        raise ValueError('series must be integer, list of integers, dictionary, or pd.Series')

    # make a dictionary with functions
    if callable(func) or func is None:
        funcdict = {a: func for a in seriesdict.keys()}
    elif isinstance(func, dict):
        funcdict = {ch: (func[ch] if ch in func.keys() else None) for ch in seriesdict.keys()}
    else:
        raise ValueError('func must be a callable, None, or a dictionary of callables (chart: callable pairs)')

    return seriesdict, funcdict





def is_leap_year(years):
    """Determine whether a year is a leap year."""
    return np.bitwise_and(years % 4 == 0,
                          np.bitwise_or(years % 100 != 0,
                                        years % 400 == 0)
                          )


def parseQuarterYear(txt: str):
    """
        Parses a string of form 'trimestre 2/2014' into '2014/6'
    Parameters
    ----------
    txt :   A string with text and two integers, one representing the year and the other the quarter (1 to 4)

    Returns
    -------
        A string representing the last month of the quarter, in yyyy/mm format.
    """
    q0, year0 = [int(x) for x in re.findall("[-+]?\d+[\.]?\d*", txt)]
    if q0 not in range(1, 5):
        if year0 in range(1, 5):
            q0, year0 = year0, q0
        else:
            raise ValueError('Cannot identify the quarter')
    return '%d/%d' % (year0, 3*int(q0))


def parseMonthYear(txt: str):
    """
        Parses a string of form 'trimestre 2/2014' into '2014/6'
    Parameters
    ----------
    txt :   A string with text and two integers, one representing the year and the other the quarter (1 to 4)

    Returns
    -------
        A string representing the last month of the quarter, in yyyy/mm format.
    """
    month0 = MONTHS.apply(lambda x: bool(re.match(x, txt, re.IGNORECASE))).idxmax()
    year0 = re.findall("[-+]?\d+[\.]?\d*", txt)[0]
    return year0 + '/' + str(month0)


MONTHS = pd.Series(['Enero', 'Febrero','Marzo', 'Abril', 'Mayo', 'Junio',
                    'Julio', 'Agosto', 'Septiembre', 'Octubre', 'Noviembre', 'Diciembre'],
                   index=range(1,13))

FREQS = pd.Series(range(4), index=['A', 'Q', 'M', 'D'])  # lower to higher frequencies


def lowestFrequency(freqs):
    original_frequencies = set([f[0].upper() for f in freqs])
    return FREQS[original_frequencies].argmin()


def findColumnTitles(data: pd.DataFrame):
    values = np.all(data.notnull().values, 1)
    return np.where(values)[0].min()

def parseDay(dia: str):
    '''
    Convert Spanish date to English date
    Parameters
    ----------
    dia String with Spanish date, format 'dd Mmm yyyy'

    Returns
    -------
    String with English date, format 'yyyy/mm/dd'
    '''
    meses = {'ene': '01', 'feb': '02', 'mar': '03', 'abr': '04', 'may': '05', 'jun': '06',
            'jul': '07', 'ago': '08', 'set': '09', 'sep': '09', 'oct': '10', 'nov': '11', 'dic': '12'}

    d, m, y = dia.split()
    return y + '/' + meses[m[:3].lower()] + '/' + d


def columns_rename(db: pd.DataFrame):
    """
    Dictionary to rename columns
    Parameters
    ----------
    db pandas DataFrame

    Returns
    -------
    A string, code to make a dictionary to rename the columns of given dataframe
    """
    print('componentes = {')
    for x in db:
        print("'" + x + "': '',")
    print('}')


def parse_date_parameter(fecha, *, inicio=True, año_primero=False):
    err_msg = """Formato de fecha no válido! Utilice:
    "yyyy/mm/dd" o  "dd/mm/yyyy" para indicar días
    "yyyy/mm" o "mm/yyyy" para indicar meses
    "yyyy" o yyyy (como int) para indicar una año
    """

    componentes = [x for x in re.split('\D', str(fecha)) if x]
    tamaños = [len(a) == 4 for a in componentes]

    if sum(tamaños) != 1 and not(tamaños[0] or tamaños[-1]):
        raise Exception(err_msg)  # debe haber un solo año, y debe aparecer al inicio o al final

    if tamaños[-1]:  # año aparece de último
        componentes.reverse()

    componentes = [int(x) for x in componentes]
    ncomp = len(componentes)



    if ncomp == 1:
        año = componentes[0]
        fecha_nueva = date(año, 1, 1) if inicio else date(año, 12, 31)
    elif ncomp == 2:
        año, mes = componentes
        fecha_nueva = date(año, mes, 1)
        if not inicio:
            fecha_nueva = (fecha_nueva + offsets.MonthEnd()).date()
    elif ncomp == 3:
        fecha_nueva = date(*componentes)
    else:
        raise Exception(err_msg)

    return fecha_nueva.strftime("%Y/%m/%d" if año_primero else "%d/%m/%Y")




def trim_data(df):
    """
    Removes missing values from start and end of series or dataframe
    Parameters
    ----------
    df: series or dataframe

    Returns
    -------
    df
    """
    return df[df.first_valid_index():df.last_valid_index()]

POSSIBLE_FREQUENCIES = {pd.Timedelta(ndias, 'D'): freq for ndias, freq in zip(
    [1, 7, 28, 29, 30, 31, 90, 91, 92, 181,182,183,184,365,366],
    ['D', 'W', 'M', 'M', 'M', 'M', 'Q', 'Q', 'Q', '6M', '6M', '6M', '6M', 'A', 'A'])}

def infer_frequency(serie):
    raw_freq = serie.index.to_series().diff().value_counts().index[0]
    return POSSIBLE_FREQUENCIES[raw_freq]


import numpy as np
import pandas as pd

from .download import downloadChart
from .scrape import loadIndicators, CHARTFREQUENCIES
from .utils import parseQuarterYear, parseMonthYear, is_leap_year, parseSeriesFreqInputs, lowestFrequency, parseDay

FIRST_OBSERVATION = {
    'YearMonth': lambda data: data.index[0] + '/01',
    'MonthYear': lambda data: data.columns[0] + '/01',
    'IndicatorYear': lambda data: data.columns[0] + '/12',
    'IndicatorQuarter': lambda data: parseQuarterYear(data.columns[0]),
    'QuarterIndicator': lambda data: parseQuarterYear(data.index[0]),
    'MonthIndicator': lambda data: parseMonthYear(data.index[0]),
    'IndicatorMonth': lambda data: parseMonthYear(data.columns[0]),
    'DayYear': lambda data: '%s/01/01' % data.columns[0],
    'DayIndicator': lambda data: parseDay(data.index[0])
}

def parse(chart, chartFormat, name=None, first=None, last=None, freq=None, func=None, quiet=True):
    #TODO: Add chartFormat='DayIndicator' to the possible options!!! Example chart=572

    data = downloadChart(chart, first, last, quiet)
    title = data._metadata['title']

    ''' CLEAN DATA'''
    if chartFormat == 'MonthYear' and 'total' in str(data.index[0]).lower():
        data.drop(data.index[0], inplace=True)
    elif chartFormat == 'DayYear':
        nonleap = np.array([~is_leap_year(int(y)) for y in data.columns])
        data.iloc[59, nonleap] = np.inf  # row 59 = Feb 29 (counting from zero-base)

    ''' GET FIRST OBSERVATION '''
    try:
        t0 = FIRST_OBSERVATION[chartFormat](data)
    except:
        raise Exception(NotImplemented)


    ''' GET DATA IN TIME v. SERIES FORMAT '''
    if chartFormat in ['IndicatorYear', 'IndicatorQuarter', 'IndicatorMonth', 'MonthYear', 'DayYear']:
        data = data.transpose()

    ''' STACK DATA '''
    if chartFormat in ['YearMonth', 'MonthYear', 'DayYear']:
        data = data.stack(dropna=False)

    ''' REMOVE 29 FEB IN NON-LEAP YEARS '''
    if chartFormat in ['DayYear']:
        data = data[data != np.inf]

    ''' REMOVE 29 REB IN NON-LEAP YEARS'''
    if chartFormat in ['DayIndicator']:
        not_a_leap_year = ~is_leap_year(np.array([int(x.split()[-1]) for x in data.index]))
        feb29 = np.array(['29 Feb' in x for x in data.index])
        wrong_day = feb29 & not_a_leap_year
        data = data.iloc[~wrong_day]

    ''' ADD TIME INDEX '''
    data.index = pd.period_range(start=t0, periods=data.shape[0], freq=CHARTFREQUENCIES[chartFormat][0])

    ''' RENAME TIME SERIES AND DROP MISSING VALUES '''
    if isinstance(data, pd.Series):
        data.name = name if name else title
        data.dropna(inplace=True)
    elif isinstance(data, pd.DataFrame):
        data.columns = [name + txt for txt in data.columns]
        data = data.dropna(0, 'all').dropna(1, 'all')
    else:
        raise ValueError('Unexpected data type: %s' % type(data))

    ''' RESAMPLE DATA '''
    if freq:
        func = func if func else np.mean
        data = data.resample(freq).apply(func)

    return data


def read(series, first=None, last=None, freq=None, func=None, quiet=True):
    """
        Reads BCCR charts.
        * This function allows downloading of charts of different formats in a single call.
        * If no frequency parameter is specified, then it will return the lowest frequency
        found in the data sets.
         * If no func parameter is specified, then higher-to lower conversion is done by taking the
         average of the data. By using a dictionary, user can specify different conversion methods for
         different charts.

    Parameters
    ----------
    series  : Charts to be downloaded, either an integer or a {int: str} dictionary.
    first   : The first year to download (integer, default=None).
    last    : The last year to download (integer, default=None)
    freq    : Data frequency (string, default=None).
    func    : How to summarize data in lower frequency (function, default=np.mean)
    quiet   : Print download info if False, nothing if True

    Returns
    -------
        Requested data, either as a pandas series (if series is int) or dataframe (if series is dict)


    Examples
    --------
        1. Download data on money supply (chart 125)

        >>> read(125)

        2. Download data on money supply (chart 125) and current account deposits on colones chart (138) since 1999. Rename
        the series as 'M1' and 'DCCMN', respectively.

        >>> read({125: 'M1', 138: 'DCCMN'}, first=1999)

        3. Same as previous example, but returning quarterly data (average of monthly observations)

        >>> read({125: 'M1', 138: 'DCCMN'}, first=1999, freq='Q')

        4. Download the consumer price index (chart 9), tradable prices (290) and non-tradable prices (289)

        >>> read(9)

        2. Download the CPI, and its tradable and non-tradable components, since 1999.

        >>> indices = {9: 'IPC', 289: 'IPC No transable', 290: 'IPC transable'}
        >>> read(indices, first=1999)








        Once downloaded, the resulting pandas dataframe can be used for plots. The first plot combines all series in a
        single plot, the latter displays a subplot for each series.

        >>> data = read({125: 'M1', 138: 'DCCMN'})
        >>> data.plot()
        >>> data.plot(subplots=True)

    Examples
    --------
        1. Download the consumer price index

        >>> read(9)

        2. Download the CPI, and its tradable and non-tradable components, since 1999.

        >>> indices = {9: 'IPC', 289: 'IPC No transable', 290: 'IPC transable'}
        >>> read(indices, first=1999)

        3. Same as before, but returning quarterly data by taking the average of monthly data.



    Examples
    --------
        1. Download data on FDI in Costa Rica, by country of origin

        >>> readIndicatorYear(2185)

        2. Download data on national accounts, constant prices

        >>> readIndicatorYear(189)

        3. Download data on national accounts, constant and current prices. Use *Real_* and *Nominal_* to tell them apart.

        >>> readIndicatorYear({189: 'Real_', 230: 'Nominal_'})

    Examples
    --------
        1. Download data on number of workers by economic activity

        >>> readIndicatorQuarter(1912)

        2. Download quarterly data on national accounts, constant prices, by industry

        >>> readIndicatorQuarter(64)

        3. Download quarterly data on national accounts, constant and current prices.
        Use *Real_* and *Nominal_* to tell them apart.

        >>> readIndicatorQuarter({68: 'Real_', 70: 'Nominal_'})

    Examples
    --------
        1. Download data on number of workers by economic activity

        >>> readIndicatorQuarter(1912)  # FIXME UPDATE DOCUMENTATION

        2. Download quarterly data on national accounts, constant prices, by industry

        >>> readIndicatorQuarter(64)

        3. Download quarterly data on national accounts, constant and current prices.
        Use *Real_* and *Nominal_* to tell them apart.

        >>> readIndicatorQuarter({68: 'Real_', 70: 'Nominal_'})

    Examples
    --------
        1. Download data on number of workers by economic activity

        >>> readIndicatorQuarter(1912)  # FIXME UPDATE DOCUMENTATION

        2. Download quarterly data on national accounts, constant prices, by industry

        >>> readIndicatorQuarter(64)

        3. Download quarterly data on national accounts, constant and current prices.
        Use *Real_* and *Nominal_* to tell them apart.

        >>> readIndicatorQuarter({68: 'Real_', 70: 'Nominal_'})

    Examples
    --------
        1. Download the Tasa basica (interest rate)

        >>> readDayYear(17)

        2. Download the Tasa basica (chart 17) and the exchange rate (chart 367), since 2000. Assign short names to series.

        >>> series = {17: 'tbp', 367: 'xr'}
        >>> readDayYear(series, first=2000)

        3. Same series as before, but returning quarterly data by taking the average of daily data.

        >>> data = readDayYear(series, freq='Q')

        Once downloaded, the resulting pandas dataframe can be used for plots. Here, we plot data from September 2006 to
        December 2009, showing each time series in a subplot.

        >>> data.plot()
        >>> data['2006-9':'2009-12'].plot(subplots=True)


    """

    seriesDict, funcdict = parseSeriesFreqInputs(series, func)
    indicators = loadIndicators()['chartFormat'][list(seriesDict.keys())]

    if freq is None:
        original_frequencies = indicators.map(CHARTFREQUENCIES)
        if original_frequencies.unique().size > 1:
            freq = lowestFrequency(original_frequencies)
            print('\nDownloading data with frequencies : %s !!!' % original_frequencies.unique())
            print('Returning frequency: %s\n' % freq)

    return pd.concat(
        [parse(chart, indicators[chart], name, first, last, freq, funcdict[chart], quiet) for chart, name in seriesDict.items()],
        axis=1)


"""
gee: Un módulo para definir la clase ServicioWeb

Este módulo defile la clase ServicioWeb y crea una instancia de la misma, SW. Esta clase permite descargar datos
del ServicioWeb del Banco Central de Costa Rica (https://www.bccr.fi.cr/indicadores-economicos/servicio-web) .

La forma usual de utilizar esta clase es

    >>> from bccr import SW

    * para buscar el código de algún indicador de interés, utilice

    >>> SW.buscar("nombre de un indicador")

    * para saber más detalles del indicador 3541 (por ejemplo)

    >>> SW.quien(3541)

    * para buscar las subcuentas de un indicador, digamos el 779

    >>> SW.subcuentas(779)

    * para descargar datos de indicadors 3, 4 y 22 (por ejemplo), hay varias formas de hacerlo

    >>> SW(indicA=4, indicB=7, indicC=231) # pasando los códigos como valores de parámetros, en
            cuyo caso los indicadores son renombrados como 'indicA', 'indicB' y 'indicC', respectivamente.
    >>> SW(3, 4, 22)   # pasando los códigos directamente (no recomendado)
"""



import pandas as pd
import numpy as np
from anytree import Node, RenderTree
import requests
import os
from dataclasses import dataclass, field
from datetime import datetime
from bs4 import BeautifulSoup
from numpy import nan
import re
from .utils import parse_date_parameter, infer_frequency

import warnings

BCCR_FOLDER = os.path.dirname(os.path.abspath(__file__))
DATA_FOLDER = os.path.join(BCCR_FOLDER, 'data')
PICKLE_FILE = os.path.join(DATA_FOLDER, 'indicadores.pkl')
EXCEL_FILE =  os.path.join(DATA_FOLDER, 'Indicadores.xlsx')

FRASE_AYUDA = """
CLASE ServicioWeb

Esta clase permite buscar y descargar datos de indicadores del servicio web del Banco Central de Costa Rica.

Suponiendo que el objeto de clase ServicioWeb se llama "SW":

    * para buscar indicadores, utilice 
        SW.buscar("nombre de un indicador")
    * para saber más detalles del indicador 8 (por ejemplo)
        SW.quien(8)
    * para buscar las subcuentas de un indicador, digamos el 784
        SW.subcuentas(784)      
    * para descargar datos de indicadors 4, 7 y 231 (por ejemplo), hay varias formas de hacerlo 
        SW(4, 7, 231)   # pasando los códigos directamente
        SW(indicA=4, indicB=7, indicC=231) # pasando los códigos como valores de parámetros, en 
            cuyo caso los indicadores son renombrados como 'indicA', 'indicB' y 'indicC', respectivamente.        
    
"""




FUNCS = {
    'mean': lambda df: np.mean(df.values),
    'sum': lambda df: np.sum(df.values),
    'last': lambda df: df.iloc[-1],
    'first': lambda df: df.iloc[0],
    'nanmean': np.nanmean,
    'nansum': np.nansum,
    'nanlast': lambda df: df.loc[df.last_valid_index()],
    'nanfirst': lambda df: df.loc[df.first_valid_index()],
}




#TODO: Arreglar problema de fechas duplicadas! ejemplo monex 3223 en 2010
#TODO: Arreglar missing values mal codificados! ejemplo monex 3223 aparecen 0s en vez de missing values


@dataclass
class ServicioWeb:
    """
    Una clase para descargar datos del servicio web del Banco Central de Costa Rica

    Attributes
    ----------
    nombre : str
        El nombre del usuario registrado en el servicio web del BCCR
    correo : str
        El correo electrónico registrado por el usuario en el servicio web.
    token : str
        El token recibido por el usuario de parte del BCCR para tener acceso al servicio web.
    indicadores : pd.DataFrame
        una table con la descripción de los indicadores disponibles en el servicio web.

    Examples
    --------
    Asumiendo que Ratón Pérez se registró en el servicio web del BCCR:

    >>> from bccr import ServicioWeb
    >>> consulta = ServicioWeb('Ratón Pérez', 'raton.perez@correo.com', '4SDLJUHKEZ')

    Si Ratón Pérez no se ha registrado en el servicio web del BCCR, entonces simplemente puede usar una instancia
    predeterminada de `ServicioWeb`, llamada `SW`, que utiliza las credenciales del paquete:

    >>> from bccr import SW

    """
    nombre: str
    correo: str
    token: str
    indicadores: pd.DataFrame

    def __usuario__(self):
        """Credenciales de usuario

        Para uso interno de la clase, en la función __descargar__.

        Returns
        -------
        dict
            el nombre, correo, y token en un diccionario con keywords Nombre, CorreoElectronico, y Token
        """
        return dict(Nombre=self.nombre, CorreoElectronico=self.correo, Token=self.token)

    def __observacion__(self, obs):
        """Extraer código de variable, fecha y valor de una observación

        Toma un elemento XML leído por BeautifulSoup y extrae los campos necesarios para identificar una observación.
        Para uso interno de la clase, en la función __descargar__.

        Parameters
        ----------
        obs: nodo de BeautifulSoup
            Representa una observación extraída de un archivo XML descargado del BCCR.

        Returns
        -------
        codigo : str
            el código del indicador
        fecha : str
            la fecha de la observación
        valor : float
            el valor de la observación
        """
        CODIGO = obs.find('COD_INDICADORINTERNO').text
        FECHA = obs.find('DES_FECHA').text[:10]
        VALOR = float(obs.find('NUM_VALOR').text) if obs.find('NUM_VALOR') else nan
        return CODIGO, FECHA, VALOR

    def __descargar__(self, Indicador, FechaInicio=None, FechaFinal=None):
        """Descargar datos del Servicio Web del BCCR

        Construye una consulta por método GET a partir de los parámetros proporcionados. Descarga los datos
        y los transforma en una tabla de datos de Pandas.
        Para uso interno de la clase, para la función datos.

        Parameters
        ----------
        Indicador : str or int
            número del indicador a descargar.
        FechaInicio : int or str, optional
            fecha de la primera observación a descargar,
            int -> año, o str -> yyyy/mm/dd (valor predeterminado es '1990/01/01')
        FechaFinal : str, optional
            fecha de la última observación a descargar, formato dd/mm/yyyy (valor predeterminado es fecha del sistema)

        Returns
        -------
        pd.Series
            Los datos solicitados, indexados por la fecha reportada por el BCCR
        """
        params = self.__usuario__()
        params['Indicador'] = Indicador
        params['FechaInicio'] = parse_date_parameter(FechaInicio, inicio=True, año_primero=False) if FechaInicio else '01/01/1900'
        params['FechaFinal'] = parse_date_parameter(FechaFinal, inicio=False, año_primero=False) if FechaFinal else datetime.now().strftime(
            '%d/%m/%Y')
        params['SubNiveles'] = 'N'
        print(params)

        host = 'https://gee.bccr.fi.cr/Indicadores/Suscripciones/WS/wsindicadoreseconomicos.asmx/ObtenerIndicadoresEconomicos'
        resp = requests.get(host, params)
        error_msg = f'\nNo se obtuvieron datos de indicador {Indicador}. Servidor respondio con mensaje {resp.reason}\n'
        error_msg += "Revise que este indicador efectivamente existe, o intente de nuevo la obtención de los datos."


        if resp.status_code == 200:  # datos recibidos exitosamente
            rawdata = resp.text
            soup = BeautifulSoup(rawdata, 'xml')
            observaciones = soup.find_all('INGC011_CAT_INDICADORECONOMIC')
            if observaciones:
                datos = [self.__observacion__(y) for y in observaciones]
                datos = pd.DataFrame(datos, columns=['variable', 'fecha', 'valor'])

                datos.index = pd.to_datetime(datos.fecha, format="%Y-%m-%d")
                return datos['valor']

            print(error_msg)
        else:
            print(error_msg)
        return None

    def __buscar_frase__(self, frase):
        """Busca indicadores cuya descripción contenga la frase indicada

        Parameters
        ----------
        frase : str
            texto que debe estar presente en la descripción del indicador

        Returns
        -------
        pd.DataFrame
            listado de indicadores que satisfacen la condición
        """
        CAMPOS = ['DESCRIPCION', 'descripcion']
        return pd.DataFrame([self.indicadores[campo].str.contains(frase, case=False) for campo in CAMPOS]).any()

    def buscar(self, todos=None, *, frase=None, algunos=None, frecuencia=None, Unidad=None, Medida=None, periodo=None):

        """buscar códigos de indicadores según su descripción

        Parameters
        ----------
        frase, todos, algunos : str
            texto que debe aparecer en la descripción del indicador. Sólo una de estos tres parámetros debe utilizarse
            a la vez. 'frase' busca una coincidencia exacta, 'todos' que todas las palabras aparezcan (en cualquier
            orden), 'algunos' que al menos una de las palabras aparezca. La búsqueda no es sensible a mayúscula/minúscula.
            Si no se indica un parámetro [por ejemplo, SW.buscar('precios consumidor')], se asume que
            SW.buscar(todos='precio consumidor')
        frecuencia : str, optional. uno de ('A','6M','Q','M','W','D')
            mostrar solo indicadores que tengan la frecuencia indicada.
        Unidad : str, optional
            mostrar solo indicadores que tengan la unidad indicada
        Medida : str, optional
            mostrar solo indicadores que tengan la medida indicada
        periodo: str, optional
            mostrar solo indicadores que tengan la periodicidad indicada

        Returns
        -------
        pd.DataFrame

        Examples
        --------

        Para buscar un indicador que tenga todas las palabras "IMAE", "tendencia", "ciclo" (en cualquier orden)

        >>> from bccr import SW
        >>> SW.buscar("IMAE tendencia ciclo")

        Para buscar un indicador que tenga la frase exacta "Índice de Precios al Consumidor"

        >>> SW.buscar(frase="Índice de Precios al Consumidor")

        Para buscar un indicador que tenga al menos una de las palabras "exportaciones" e "importaciones"

        >>> SW.buscar(algunos="exportaciones importaciones")

        Para buscar los términos "precios transables", y filtrar que muestre solo resultados con medida "Variación interanual"

        >>> SW.buscar('precios transables', Medida='Variación interanual')

        Para mostrar un breve mensaje de ayuda

        >>> SW.buscar()


        Notes
        -----
        El parámetro 'frecuencia' será discontinuado en algún momento. Se recomienda usar el parámetro 'periodo' para
        cumplir la misma funcionalidad.
        """
        CAMPOS = ['DESCRIPCION', 'descripcion', 'Unidad','Medida','periodo']

        if frase:
            temp = self.__buscar_frase__(frase)
        elif todos:
            temp = pd.DataFrame([self.__buscar_frase__(palabra) for palabra in todos.split(' ')]).all()
        elif algunos:
            temp = pd.DataFrame([self.__buscar_frase__(palabra) for palabra in algunos.split(' ')]).any()
        else:
            ayuda = """ BUSCAR
            Esta función ayuda a buscar los códigos de indicadores, utilizando palabras descriptivas.
            Exactamente un parámetro de [frase, todos, algunos] debe ser proporcionado.
            
            Ejemplos de uso:
                buscar("descripción contiene todos estos términos en cualquir orden")
                buscar(frase="descripción contiene esta frase literalmente")
                buscar(algunos="descripción contiene alguno de estos términos")
                buscar()  # muestra este mensaje de ayuda
            """
            print(ayuda)
            return
        results = self.indicadores[temp].copy()

        if frecuencia:
            freq = frecuencia[0].upper()
            results.query('freq == @freq', inplace=True)
        if Unidad:
            results.query('Unidad == @Unidad', inplace=True)
        if Medida:
            results.query('Medida == @Medida', inplace=True)
        if periodo:
            results.query('periodo == @periodo', inplace=True)

        return results[CAMPOS]


    def __actualizar_catalogo__(self):
        """Actualiza el catálago de indicadores

        Returns
        -------
        pd.DataFrame
            el catálogo actualizado

        Notes
        -----
        Esta función es  para uso interno del paquete. Asume la existencia de un archivo de Excel con el catálogo, e
        intenta añadirlo al paquete para futura referencia.

        """
        FREQ = {'Anual': 'A',
                'Mensual': 'M',
                'Trimestral': 'Q',
                'Nueva semanal': 'W',
                'Diaria': 'D',
                'Semestral': '6M',
                'Semanal': 'W'}


        CAMPOS = {'INGC011_COD_INDICADORECONOMIC': 'codigo',
                  'INGC011_COD_INDICADORINTERNO': 'cuenta',
                  'INGC011_NOM_INDICECONOMICOESP': 'nombre',
                  'INGC011_DES_TITULOESPANOL': 'descripcion',
                  'INGC012_COD_MEDIDA': 'medida',
                  'INGC025_COD_UNIDAD': 'unidad',
                  'Periodicidad': 'periodo'}

        unidades = pd.read_excel(EXCEL_FILE, 'INGC012_COD_MEDIDA', dtype=str)
        unidades = unidades.set_index('CodUnidad')['NomUnidadespanol']
        medidas = pd.read_excel(EXCEL_FILE, 'INGC025_COD_UNIDAD', dtype=str)
        medidas = medidas.set_index('Codigo')['NombreEspannol']

        indicadores = pd.read_excel(EXCEL_FILE, 'Indicadores', dtype=str).rename(columns=CAMPOS)
        indicadores['Unidad'] = indicadores['unidad'].map(unidades)
        indicadores['Medida'] = indicadores['medida'].map(medidas)
        indicadores['freq'] = indicadores['periodo'].map(pd.Series(FREQ))
        indicadores['freq'] = indicadores['freq'].astype('category').cat.set_categories(['A', '6M', 'Q', 'M', 'W', 'D'], ordered=True)

        for campo in indicadores:
            indicadores[campo] = indicadores[campo].str.strip()

        indicadores.set_index('codigo', inplace=True)
        indicadores['familia'] = indicadores['cuenta'].str[28:]


        FAMILIAS = indicadores['familia'].value_counts()
        todo = pd.concat([self.__hacer_arbol__(ff, indicadores) for ff in FAMILIAS.index], keys=FAMILIAS.index)
        todo.index = todo.index.get_level_values(1) + '.' + todo.index.get_level_values(0)
        todo.sort_index(inplace=True)

        indicadores = indicadores.merge(todo[['node']], left_on='cuenta', right_index=True)

        indicadores['DESCRIPCION'] = indicadores['node'].astype(str)

        self.indicadores = indicadores
        try:
            indicadores.to_pickle(PICKLE_FILE)
        except:
            print('La tabla indicadores fue actualizada en el objeto, pero no en el paquete bccr.')

    def __hacer_arbol__(self, familia, indicadores):
        """Representa un indicador como parte de un subconjunto de indicadores

        Parameters
        ----------
        familia : str
            código de un indicador en el esquema de cuentas
        indicadores : pd.DataFrame
            tabla con el catálogo de indicadores

        Returns
        -------
        pd.DataFrame
            listado de las cuentas
        """
        grupos = pd.DataFrame(
            dict(nombre=['BCCR', 'Índices de Precios', 'Tipos de cambio', 'Tasas de interés', 'Sector Real',
                         'Sector Externo', 'Sector Fiscal', 'Sector Monetario y Financiero',
                         'Mercados de Negociación', 'Demográficas y Mercado Laboral', 'Expectativas'],
                 codigo=['***'] * 11,
                 cuenta=[f'E{k:02}.00.00.00.00.00.00.00.00' for k in range(11)]))

        def find_parent(cta, catalogo):
            subctas = cta.split('.')
            k = ((np.array(subctas) != "00") * np.arange(9)).max()
            subctas[k] = '00' if k else 'E00'
            cta1 = '.'.join(subctas)
            return cta1 if cta1 in catalogo.index else find_parent(cta1, catalogo)


        filas = indicadores['cuenta'].str.endswith(familia)
        ind = indicadores[filas].reset_index()[['nombre', 'codigo', 'cuenta']]
        ind['cuenta'] = ind['cuenta'].str[:-len(familia) - 1]
        ind = ind.append(grupos, sort=True)
        ind.index = ind['cuenta']
        ind.sort_index(inplace=True)
        ind['parent'] = ind['cuenta'].apply(find_parent, args=(ind['cuenta'],))
        ind.loc['E00.00.00.00.00.00.00.00.00', 'node'] = Node('BCCR')
        for cuenta, (nombre, parent, codigo) in ind[['nombre', 'parent', 'codigo']].iloc[1:].iterrows():
            cdg = '' if codigo == '***' else f' [{codigo}]'
            ind.loc[cuenta, 'node'] = Node(name=nombre + cdg, parent=ind.loc[parent, 'node'])
        ind.drop(grupos['cuenta'], inplace=True)
        return ind


    def __print_node__(self,node):
        """Imprime un árbol que representa la posición de un nodo en el esquema de cuentas.

        Para uso interno de la clase.

        Parameters
        ----------
        node: Node (anytree)
            indicador que se quiere representar, como un nodo del catálogo de indicadores
        Returns
        -------
        None :
            El resultado se imprime a pantalla
        """
        lista = str(node)[12:-1].split('/')
        for i, n in enumerate(lista):
            print('|' + '-' * (3 * (i + 1)) + ' ' + n)

    def quien(self, codigo):
        """Imprime información acerca de un indicador

        Parameters
        ----------
        codigo : str or int
            código numérico del indicador que se desea describir.

        Returns
        -------
        None :
            El resultado se imprime a pantalla

        Examples
        --------
        >>> from bccr import SW
        >>> SW.quien(33438)
        """
        if str(codigo) not in self.indicadores.index:
            print(f'La variable {codigo} no aparece en la lista de indicadores.')
            print('Puede ser que esta variable sí exista en la base de datos del BCCR.')
            return

        A = self.indicadores.loc[str(codigo)]

        ind = self.__hacer_arbol__(A['cuenta'][28:], self.indicadores)
        print(f"Variable {codigo} >>>")
        print(f"   Nombre      : {A['nombre']}.")
        print(f"   Descripcion : {A['descripcion']}.")
        print(f"   Unidad      : {A['Unidad']}" + (f" ({A['Medida']})." if int(A['medida']) > 2 else "."))
        print(f"   Periodicidad: {A['periodo']}.\n")
        self.__print_node__(ind.loc[A['cuenta'][:27]].node)
        print('\n')


    def __nombre__(self, codigo):
        """
        Nombre del indicador con ese código en el catálogo. Si no existe, entonces el código mismo
        Parameters
        ----------
        codigo int or str, código del indicador

        Returns
        -------
        str
        """
        codigo = str(codigo)
        return self.indicadores.loc[codigo, 'nombre'] if codigo in self.indicadores.index else codigo

    def __frecuencia__(self, codigo, serie):
        """
        Frecuencia del indicador codigo. Si existe en el catálogo del paquete, usa el catálogo. De lo contrario, lo
        infiere a partir del índice de los datos
        Parameters
        ----------
        codigo int or str, código del indicador
        serie pd.Series, datos indexados como serie de tiempo para inferir su frecuencia.

        Returns
        -------
            str
        """
        if codigo in self.indicadores.index:
            return self.indicadores.loc[codigo, 'freq']
        else:
            return infer_frequency(serie)


    def subcuentas(self, codigo):
        """Subcuentas de un indicador

        Algunos indicadores pueden desagregarse (por ejemplo, el IMAE total se puede desagregar por actividad económica).
        Esta función ayuda a encontrar los códigos de esas subcuentas.

        Parameters
        ----------
        codigo : str or int
            código numérico del indicador del que se desea conocer sus subcuentas.

        Returns
        -------
        list:
            Los códigos de las subcuentas.
            Además, el resultado se imprime a pantalla como un árbol de cuentas.

        Examples
        --------
        >>> from bccr import SW
        >>> SW.subcuentas(33439)
        """
        cta = self.indicadores.loc[str(codigo), 'node']
        treestr = RenderTree(cta).by_attr()
        print(treestr)
        return re.findall('\[([0-9]+)\]', treestr)

    def datos(self, *codigos, FechaInicio=None, FechaFinal=None, func=None, freq=None, fillna=None, **indicadores):
        """

        Parameters
        ----------
        codigos: sucesión de enteros o strings
                Los códigos numéricos (como int o str) de los indicadores que se desean descargar (ver parámetro
                `indicadores`). En el resultado, el indicador se identifica con el nombre que tiene en el catálogo de
                cuentas; en caso de querer indicar un nombre distinto, solicitar el indicador usando el parámetro
                `indicadores`.
        FechaInicio : str or int, optional
            fecha de la primera observación a descargar. Ver nota 6 abajo para más detalles.
        FechaFinal : str or int, optional
            fecha de la última observación a descargar. Ver nota 6 abajo para más detalles.
        func : str
            Una opción de ['mean', 'sum', 'last', 'first', 'nanmean', 'nansum', 'nanlast', 'nanfirst'].
            función que se desea utilizar para transformar la frecuencia de los datos (predeterminado: None).
            Si se especifica una sola función, se usa la misma para todas las series que lo requiera. Para usar funciones
            distintas puede pasar un diccionario {nombre-de-la-serie: funcion-a-usar,...}
        freq : str, optional
            frecuencia a la que se quiere convertir los datos. Si este parámetro no se indica y se requieren series de
            distintas frecuencias, el resultado será un Data.Frame con la menor periodicidad de las series solicitadas.
        fillna : str, one of ('backfill', 'bfill', 'pad', 'ffill').
            Si hay datos faltantes (por ejemplo, fines de semana), cómo rellenar esos valores. 'backfill' y 'bfill' usan
             el siguiente valor disponible (llena hacia atrás) mientras que 'pad' y 'ffill' usan el último valor
             disponible (llena hacia adelante).
        indicadores: pares de nombre=codigo, de las series requeridas
             Este es el método preferible para indicar las series requeridas (en vez del parámetro `códigos`). Las series
             se enumeran como pares de nombre=codigo, donde nombre es el nombre que tendrá el indicador en el DataFrame
             resultante, y codigo es un número entero que identifica la serie en el servicio web del BCCR.

        Returns
        -------
        pd.DataFrame
            una tabla con los datos solicitados, filas indexadas por tiempo, cada columna es un indicador.

        Notes
        -----
        1. Esta función construye una consulta por método GET a partir de los parámetros proporcionados, para cada uno de
        los Indicadores solicitados. Descarga los datos, los transforma en una tabla de datos de Pandas.

        2. Si hay indicadores de distintas frecuecias, los transforma a la misma frecuencia según el método indicado.

        3. A excepción de `codigos`, todos los parámetros deben usar palabra clave, aunque es preferible usar `indicadores`
        en vez de `codigos`. (Ver los ejemplos)

        4. Hay varias maneras de indicar los Indicadores: (a) simplemente enumerándolos, (b) en una lista o tupla, y
        (c) en un diccionario (método obsoleto, es mejor usar el parámetro `indicadores`). En caso de tratarse de
        diccionario, los códigos se indican como las llaves del diccionario, mientras que los valores del diccionario
        se usan para renombrar las columnas resultantes.

        5. Las instancias de la clase ServicioWeb son ejecutables: si se llaman como una función, simplemente ejecutan la
        función datos()

        6. El formato de fechas es muy flexible. Por ejemplo, todas estas expresiones son válidas:

        - Para indicar el año 2015: se puede emplear tanto un `int` como un `str`

        >>>   FechaInicio = 2015 # se interpreta como  1 de enero de 2015
        >>>   FechaFinal = "2015"  # se interpreta como 31 de diciembre de 2015

        - Para indicar el mes marzo de 2017: cualquiera de estas expresiones es válida

         >>> "2017-03"
         >>> "2017/03"
         >>> "2017m3"
         >>> "03/2017"
         >>> "03-2017"

         En `FechaInicio=` resulta en 1 de marzo de 2017, en `FechaFinal=` resulta en 31 de marzo de 2017.

        - Para indicar el 12 de agosto de 2018:

         >>> "2017/8/12"
         >>> "2017-08-12"
         >>> "12/8/2017"

        Observe que para separar los componentes de una fecha se puede usar cualquier caracter no numérico (usualmente `/` o `-`).

        Examples
        --------
        La forma preferible de solicitar los indicadores es como pares de `nombre=código`, de manera que `nombre` se utilice
        como encabezado de columna en la tabla de datos resultante:

        >>> from bccr import SW
        >>> SW(IMAE=35449, Inflación=25485)

        Para hacer consultas rápidas, puede simplemente pasarse únicamente los códigos

        >>> SW(35449, 25485)

        Observe que si se "ejecuta" la instancia `SW`, simplemente se está ejecutando la función `datos`. Por ello, no es
        necesario escribir `datos` en los ejemplos anteriores

        >>> SW.datos(35449, 25485) # resultado idéntico al anterior, pero digitando más.

        Si se desea, los códigos pueden escribirse como de tipo `str`:

        >>> SW('35449', '25485')

        Puede también pasarse los códigos en un diccionario. **Advertencia** Esta forma se considera obsoleta, pero se mantiene
        aún en el paquete para darle soporte a código escrito con versiones anteriores de `bccr`. Nótese que en esta forma, las
        llaves del diccionario son los códigos, y los valores se usan como encabezado de columna:

        >>> cuentas = {33439:'PIB', 33448:'Consumo', 33451:'Gasto', 33457:'Inversión'}
        >>> SW(cuentas)    # esta forma es obsoleta, será eliminada en una versión futura

        Si desea usar diccionarios, la forma preferible de hacer la consulta anterior es (observe cambio de orden en diccionario
        y que debe desempacarlo con dos asteriscos ** ):

        >>> cuentas = {'PIB': 33439, 'Consumo': 33448, 'Gasto': 33451, 'Inversión':33457}
        >>> SW(**cuentas)    # es necesario usar ** delante del nombre del diccionario
        >>> SW(PIB=33439, Consumo=33448, Gasto=33451, Inversión=33457)  # equivalente a las dos líneas anteriores

        Es posible restringir el rango de los datos, usando los parámetros `FechaInicio=` y `FechaFinal=`:

        >>> SW(IMAE=35449, Inflación=25485, FechaInicio='2000/01/01')

        En caso de solo señalar el año, se sobrentiende "1 de enero" en el caso de `FechaInicio=` y "31 de diciembre" en `FechaFinal=`

        >>> SW(IMAE=35449, Inflación=25485, FechaInicio=2000, FechaFinal=2015)

        Si se solicitan indicadores que tienen distinta periodicidad, el resultado tendrá la periodicidad del indicador menos frecuente.
        Para ello, se puede indicar la función para convertir los datos de mayor a menor frecuencia, y si de seben rellenar datos
        faltantes (por ejemplo, valores faltantes en los fines de semana se pueden intrapolar con el dato del lunes siguiente o
        del viernes anterior)

        >>> SW(IMAE=35449, Inflación=25485, TPM=3541, func='mean', fillna='ffill') # IMAE e Inflación son series mensuales,
        >>> # miestras que TPM es diaria. TPM se convierte en mensual calculando el promedio, habiendo sustituido los
        >>> # valores faltantes con el último dato disponible
        """

        msg = """
        En una futura versión, los indicadores deberán ser solicitados como pares de 'nombre=codigo', o bien 
        como un listado de enteros. Por ejemplo
            SW(TPM=3541, IMAE=913, Inflación=25485)
        o bien
            SW(3541, 913, 25485)                
        """
        for indic in codigos:
            if isinstance(indic, dict):
                warnings.warn(msg)
                for key, value in indic.items():
                    indicadores[value] = key
            elif isinstance(indic, str) or isinstance(indic, int):
                indicadores[self.__nombre__(indic)] = indic
            elif hasattr(indic, '__iter__'):
                warnings.warn(msg)
                for val in indic:
                    indicadores[self.__nombre__(val)] = val
            else:
                raise('No sé cómo interpretar el indicador requerido')

        # Convertir numeros de cuadros a textos
        indicadores = {nombre: str(codigo)  for nombre, codigo in indicadores.items()}


        # Descargar los datos
        datos = {nombre: self.__descargar__(codigo, FechaInicio, FechaFinal) for nombre, codigo in indicadores.items()}

        # Desechar campos de indicadores no descargados
        datos = {nombre: df for nombre, df in datos.items() if df is not None}

        if len(datos) == 0:  # no se encontró ninguna serie, devolver DataFrame vacío.
            return pd.DataFrame(columns=indicadores.keys())


        # LLenar los espacios en blanco en la línea de tiemp
        freqs = {nombre: self.__frecuencia__(codigo, datos[nombre]) for nombre, codigo in indicadores.items() if nombre in datos}
        nfreqs = len(set(freqs.values()))
        datos = {nombre: serie.resample(freqs[nombre]).mean() for nombre, serie in datos.items() if serie is not None}

        if isinstance(fillna, str):
            fillna = fillna.lower()
            if fillna in ('backfill', 'bfill', 'pad', 'ffill'):
                datos = {nombre: serie.fillna(method=fillna) for nombre, serie in datos.items()}
            else:
                warnings.warn("Método de remplazo de valores faltantes debe ser uno de ('backfill', 'bfill', 'pad', 'ffill')")
        elif fillna is not None:
            warnings.warn("Método de remplazo de valores faltantes debe ser uno de ('backfill', 'bfill', 'pad', 'ffill')")


        # Convertir frecuencias, si es necesario o requerido
        if nfreqs > 1 or freq:
            freqs2 = pd.Series(freqs).astype('category').cat.set_categories(['A', '6M', 'Q', 'M', 'W', 'D'], ordered=True)
            freq = freq if freq else freqs2.min()
            if func is None:
                msg = """
                En una futura versión, si necesita cambiar la frecuencia de los datos deberá indicar la función a utilizar
                para dicha conversión, ya sea con 'func=nombre_de_funcion' (se aplica esta función a todas las variable que
                requieran cambiar de frecuencia), o bien con un diccionario con claves iguales a los nombres de indicadores
                 (o códigos, si no asigna nombres) y valores iguales a la función a utilizar para ese indicador.
                 
                Utilizando func=numpy.sum para esta conversión. 
                """
                warnings.warn(msg)
                func = np.sum

            if type(func) is str:
                if func not in FUNCS.keys():
                    raise ValueError("El parámetro 'func' debe ser uno de " + str(list(FUNCS.keys())))
                else:
                    funciones = {nombre: FUNCS[func] for nombre in indicadores}
            elif callable(func):
                funciones = {nombre: func for nombre in indicadores}
            else:
                raise ValueError("El parámetro 'func' debe ser un str, función, o diccionario")

            for nombre, serie in datos.items():
                if freqs[nombre] != freq:
                    if serie.isna().any():
                        msg = f"Se detectaron valores faltantes en serie '{nombre}' antes de cambiarle de frecuencia.\n"
                        msg += "Considere completar esos datos usando el parámetro 'fillna' de esta función."
                        warnings.warn(msg)
                    datos[nombre] = serie.resample(freq).apply(funciones[nombre])

        datos = pd.DataFrame(datos)
        datos.index = datos.index.to_period()
        return datos


    def __call__(self, *args, **kwargs):
        return self.datos(*args, **kwargs)

    def __str__(self):
        return FRASE_AYUDA

    def __repr__(self):
        return self.__str__()


SW = ServicioWeb(nombre = 'Paquete BCCR Python',
                 correo = 'paquete.bccr.python@outlook.com',
                 token = '5CMRCBTHMT',
                 indicadores = pd.read_pickle(PICKLE_FILE))


import os
import re
import requests
import pandas as pd
from bs4 import BeautifulSoup

from .download import api


CHARTFREQUENCIES = {
    'MonthYear': 'Monthly',
    'YearMonth': 'Monthly',
    'IndicatorYear': 'Annual',
    'IndicatorQuarter': 'Quarterly',
    'IndicatorMonth': 'Monthly',
    'QuarterIndicator': 'Quarterly',
    'MonthIndicator': 'Monthly',
    'DayYear': 'Daily',
    'DayIndicator': 'Daily',
    'NOT-SUPPORTED-YET': 'unknown'
}



def fastTitle(chart, quiet=True):
    """
        Read title of a single chart. Optimized for speed: try to download as little data as possible, if it fails then
        download using default dates
    Parameters
    ----------
    chart   : chart number (integer)

    Returns
    -------
            A pandas series with two elements (title and subtitle)
    """

    if not quiet:
        print(api(chart, excel=False))

    try:
        rawdata = pd.read_html(api(chart, 2015, 2015), thousands="")[0]
    except:
        rawdata = pd.read_html(api(chart), thousands="")[0]

    df = pd.DataFrame(index=[chart])
    df['title'], df['subtitle'], subts2 = rawdata.iloc[:3, 0]
    if not pd.isnull(subts2):
        df['subtitle'] += ' --- %s' % subts2

    return df


def readTitle(series, quiet=True):
    """
        Reads the title and subtitle of indicated series

    Parameters
    ----------
    series  : An iterable of integers, indicating chart numbers

    Returns
    -------
        A pandas dataframe, indexed by chart numbers

    """
    seriesSet = [series] if isinstance(series, int) else set(series)
    return pd.concat([fastTitle(v, quiet) for v in seriesSet])







def findAllCharts():
    baseURL = 'https://www.bccr.fi.cr/seccion-indicadores-economicos/'
    pages = ['mercado-negociación',
             'monetario-y-financiero',
             'producción-y-empleo',
             'sector-externo',
             'tasas-de-interés',
             'tipos-de-cambio',
             'encuestas_economicas',
             'finanzas-públicas',
             'índices-de-precios']

    urls = {page: baseURL + page for page in pages}


    pagesCharts = []
    for sector, url in urls.items():
        print('Buscando cuadros en', url)

        chartsList = list()

        access_page = requests.get(url).content
        soup = BeautifulSoup(access_page, 'lxml')
        url_list_of_indicators = soup.findAll('iframe')[0]['src']

        sector_page = requests.get(url_list_of_indicators).content
        soup = BeautifulSoup(sector_page, 'lxml')

        print('\n\nsector = ', sector)

        for kk in soup.find_all('a'):
            link = kk['href']
            if 'CodCuadro=' in link:
                chartsList.append(int(link.split('CodCuadro=')[1]))

        titlestable = readTitle(chartsList, False)
        titlestable['sector'] = sector
        print(titlestable)
        pagesCharts.append(titlestable)

    data = pd.concat(pagesCharts, axis=0)
    data['sector'] = pd.Categorical(data['sector'])

    '''SAVE THE DATA'''
    oldDir = os.getcwd()
    os.chdir(os.path.dirname(os.path.abspath(__file__)))
    pd.to_pickle(data, './data/asReadFromBCCR.pkl')
    os.chdir(oldDir)









def search(expression, match_all=True, subtitle=False):
    """
        Find indicators by (partial) name match
    Parameters
    ----------
    expression  : A string to search in chart titles (case insensitive), terms separated by spaces
    match_all   : match all terms if True, match any term if False
    subtitle    : whether to search in chart subtitle too.

    Returns
    -------
        A pandas dataframe with all charts whose titles contain a given string (case insensitive)

    Examples
    --------

        >>> search('producto')
        >>> search('tasa')
        >>> search('precio consumidor')
        >>> search('exportaciones importaciones')
        >>> search('exportaciones importaciones', False)
    """
    indicators = loadIndicators()
    titles = indicators['title'] + (indicators['subtitle'] if subtitle else '')

    tt = [titles.apply(lambda x: bool(re.search(name, x, re.IGNORECASE))) for name in expression.split()]
    tt = pd.concat(tt, axis=1)
    tt = tt.all(1) if match_all else tt.any(1)
    results = indicators.loc[tt, ['title', 'subtitle', 'chartFormat']]


    results['frequency'] = results['chartFormat'].apply(lambda x: CHARTFREQUENCIES[x].lower())

    return results[['title', 'subtitle', 'frequency']]


def loadIndicators():
    oldDir = os.getcwd()
    os.chdir(os.path.dirname(os.path.abspath(__file__)))
    indicators = pd.read_pickle('data/indicators.pkl')
    os.chdir(oldDir)
    return indicators


def dropDuplicateIndices(df):
    df.reset_index(inplace=True)
    df.drop_duplicates(subset='index', inplace=True)
    df.set_index('index', inplace=True)
    return df



def updateIndicators():
    '''
    Updates the indicators database

    This function is necessary to update the database in the following situations:
        1. To add new charts
        2. To specify a read method for a previously unsupported chart.

    Before running this function, one needs to update the ./data/allChartFormats.xlsx file with the chart number and its
        associated read method. After saving the updated Excel file, run this function so that bccr.read() can read the
        new charts.

    Returns
    -------
        A pandas dataframe with metadata for all charts in the BCCR website

    '''
    oldDir = os.getcwd()
    os.chdir(os.path.dirname(os.path.abspath(__file__)))

    ''' OPEN DATABASES, DELETE DUPLICATES'''
    webData = dropDuplicateIndices(pd.read_pickle('./data/asReadFromBCCR.pkl'))
    chartData = dropDuplicateIndices(pd.read_excel('./data/allChartFormats.xlsx'))
    indicators = pd.concat([webData, chartData], axis=1)
    indicators['chartFormat'] = pd.Categorical(indicators['chartFormat'])
    indicators.index.names = ['chart']
    pd.to_pickle(indicators, './data/indicators.pkl')
    os.chdir(oldDir)
    return indicators



from jupyter_dash import JupyterDash
import pandas as pd
from dash.dependencies import Input, Output, State
import dash_table
import dash_core_components as dcc
import dash_html_components as html
from dash_extensions import Download
from dash_extensions.snippets import send_data_frame
from dash_table.Format import Format, Scheme
import plotly.express as px
from bccr import SW
import io
from contextlib import redirect_stdout

from datetime import date

import webbrowser

DATOS_PARA_DESCARGAR = [pd.DataFrame()]


# Esta parte controla asuntos de estética de la página
external_stylesheets = ['https://codepen.io/chriddyp/pen/bWLwgP.css']



ayuda = """
# Acerca del paquete bccr

## Una API de Python para descargar datos del Banco Central de Costa Rica

El propósito de esta interfase gráfica es facilitar la búsqueda y descarga de indicadores publicados por el [Banco Central de Costa Rica](https://www.bccr.fi.cr/) por medio de su servicio web. 

La interfase está basada en el paquete de Python `bccr` desarrollada por el autor de esta interfase. El paquete ofrece dos clases para buscar datos y descargarlos:

* [`ServicioWeb`](http://randall-romero.com/demo-bccr-servicioweb/): descarga indicadores individuales del [servicio web](https://www.bccr.fi.cr/indicadores-economicos/servicio-web) del Banco Central.
* [`PaginaWeb`](http://randall-romero.com/demo-bccr-paginaweb/): descarga cuadros individuales (algunos con un solo indicador, otros con varios), de la página de [indicadores económicos](https://www.bccr.fi.cr/indicadores-economicos).

En esta versión, la interfase solo permite descargar datos a través del `Servicioweb`. 

Los datos se descargan y se presentan en una tabla en formato *tidy*, en la cual cada fila corresponde a un período (día, mes, trimestre, año) y cada columna a un indicador. 

---

## Parámetros de búsqueda (panel izquierdo):
### Indicadores
En esta tabla se indican los códigos de los indicadores que se desean descargar, así como el nombre que se le quiere dar al indicador
en la tabla de datos. Puede agregar indicadores usando el botón "Agregar otro código".

### Rango de fechas a consultar:
Delimita el rango de datos que se descargan. Las fechas se deben indicar en formato yyyy/mm/dd.

### Frecuencia de los datos:
De manera predeterminada se muestran datos en la frecuencia original (si se solicitan indicadores de distinta frecuencia, se 
reportan con la menor de las frecuencias). Se puede especificar una frecuencia menor (por ejemplo, trimestral para datos diarios o mensuales, pero no para datos anuales).

### Función para agregación 
Si se cambia la frecuencia de los datos, esta función determina cómo se agregan los datos de un período (promedio, suma, último o primero). Puede especificar si esta agregación debe propagar o ignorar los valores faltantes. 

### Rellenar datos faltantes
En algunas series es común que hayan brechas en la serie de datos (por ejemplo, en series diarias si no se reporta dato los fines
de semana). Esta opción permite imputar valores a estos datos faltantes (ya sea usando el último dato disponible o el siguiente).
Esta operación se hace antes de la agregación de datos (si se cambia la frecuencia de las series). 

---

## Viñetas
### Datos
Presenta una tabla con los datos descargados. Puede descargar estos datos oprimiendo los botones que aparecen en la parte superior.
Además, en la parte superior de la viñeta aparece el código de Python que descarga los datos mostrados en la tabla.

### Gráfico
Presenta un gráfico de todas las series descargadas. Puede ocultar y volver a mostrar series individuales haciendo clic a su
nombre en la leyenda del gráfico.

### Buscar códigos
Facilita la obtención de los códigos del catálogo de indicadores del Servicio Web. La tabla resultante de buscar un término
(palabra o frases) puede filtrarse escribiendo en la fila superior (por ejemplo, para que solo muestre indicadores como 
porcentaje de variación, o solo indicadores mensuales).

### ¿Quién?
Permite identificar las relaciones de las cuentas, mostrando las cuentas superiores así como las subcuentas, para un código
especificado. Esto sirve, por ejemplo, si de desea descargar los componentes del IPC. 

---

## Acerca del autor
Randall Romero Aguilar, PhD

[randall-romero.com](http://randall-romero.com)

Si desea reportarme un fallo en esta herramienta, o bien sugerirme alguna mejora, por favor escríbame al correo
[randall.romero@outlook.com](mailto:randall.romero@outlook.com)

---

## Aviso importante

Este paquete no es un producto oficial de BCCR. El autor lo provee para facilitar el manejo de datos, pero no ofrece ninguna garantía acerca de su correcto funcionamiento. 
"""




app = JupyterDash(__name__, external_stylesheets=external_stylesheets, prevent_initial_callbacks=True)
app.layout = html.Div([
    html.Div(
        children=[
            html.H2('Indicadores'),
            dash_table.DataTable(
                id='seleccionar-indicadores',
                columns=[
                    {'name': 'Nombre', 'id': 'nombre', 'deletable': False, 'renamable': False},
                    {'name': 'Código', 'id': 'código', 'deletable': False, 'renamable': False},
                ],
                data=[{'nombre': 'nombre_variable', 'código': ''}],
                editable=True,
                row_deletable=True
            ), # fin tabla de códigos a descargar
            html.Button('Agregar otro código', id='agregar-codigo-button', n_clicks=0),
            html.H3('Parámetros opcionales'),
            html.H4('Rango de fechas a consultar'),
            dcc.DatePickerSingle(
                id='fecha-primera-observación',
                display_format='YYYY/MM/DD',
                min_date_allowed=date(1950, 1, 1),
                max_date_allowed=date(2030,12,31),
                initial_visible_month=date(1990, 1, 1),
                date=None #date(1990, 1, 1)
            ), # fin fecha de inicio
            dcc.DatePickerSingle(
                id='fecha-última-observación',
                display_format='YYYY/MM/DD',
                min_date_allowed=date(1950, 1, 1),
                max_date_allowed=date(2030,12,31),
                initial_visible_month=date.today(), #(2021, 1, 1),
                date=None #date.today()
            ), # fin fecha de cierre
            html.H4('Frecuencia de los datos'),
            dcc.Dropdown(
                id='frecuencia',
                options=[
                    {'label': 'Original', 'value': 'O'},
                    {'label': 'Anual', 'value': 'A'},
                    {'label': 'Semestral', 'value': '6M'},
                    {'label': 'Trimestral', 'value': 'Q'},
                    {'label': 'Mensual', 'value': 'M'},
                    {'label': 'Semanal', 'value': 'W'},
                    {'label': 'Diaria', 'value': 'D'}
                ],
                value='O'
            ), # fin de Frecuencia de los datos
            html.H4('Función para agregación'),
            dcc.Dropdown(
                id='func',
                options=[
                    {'label': '  ', 'value': 'None'},
                    {'label': 'Promedio', 'value': 'mean'},
                    {'label': 'Suma', 'value': 'sum'},
                    {'label': 'Último', 'value': 'last'},
                    {'label': 'Primero', 'value': 'first'},
                    {'label': 'Promedio, ignorando NaN', 'value': 'nanmean'},
                    {'label': 'Suma, ignorando NaN', 'value': 'nansum'},
                    {'label': 'Último, ignorando NaN', 'value': 'nanlast'},
                    {'label': 'Primero, ignorando NaN', 'value': 'nanfirst'},
                ],
                value='None'
            ), # fin de función para cambiar frecuencia de los datos
            html.H4('Rellenar datos faltantes'),
            dcc.Dropdown(
                id='fillna',
                options=[
                    {'label': 'No', 'value': 'no'},
                    {'label': 'Usar último disponible', 'value': 'ffill'},
                    {'label': 'Usar próximo disponible', 'value': 'bfill'},
                ],
                value='no'
            ), # fin de función para cambiar frecuencia de los datos
            html.Button('Consultar datos', id='consultar-datos-button', n_clicks=0,
                        style={'width': '80%', 'font-size': '1.25em', 'margin': '40px',
                               'color':'white', 'background-color': 'OrangeRed'}),
            html.Img(src='https://randall-romero.github.io/econometria/_static/r2-logo.png',width='80%', style={'margin-left': '40px'}),
        ],
        style={'width': '20%', 'display': 'inline-block', 'background-color': 'AliceBlue'}
    ), # fin panel controles
    html.Div( children=[
                 dcc.Tabs([
                     # Panel 1: TABLA DE DATOS================================================================
                     dcc.Tab(label='Datos',
                             children=[
                                 dcc.Textarea(id='comando',
                                              value='Acá aparecerá el código que descarga sus datos',
                                              readOnly=True,
                                              style={'width': '100%', 'font-size': '1.5em',
                                                     'color':'white', 'background-color': 'LightSlateGray'},
                                              ),
                                 html.Table(children=[
                                     html.Tr(children=[
                                         html.Td(dcc.Markdown("Para descargar los datos ⇒")),
                                         html.Td(dcc.Input(id='file-name', type='text', value='nombre-de-archivo', size='16')),
                                         html.Td([html.Button("Download Excel", id='btn-xlsx'),
                                                  Download(id="download-dataframe-xlsx")]),
                                         html.Td([html.Button("Download Stata", id='btn-dta'),
                                                  Download(id="download-dataframe-dta")]),
                                         html.Td([html.Button("Download CSV", id='btn-csv'),
                                                  Download(id="download-dataframe-csv")]),
                                     ])
                                 ]),
                                 dash_table.DataTable(
                                     id='datos-descargados',
                                     editable=False,
                                     row_deletable=False,
                                     style_as_list_view=True,
                                 )
                             ],
                             ),
                     # PANEL 2:  GRÁFICO======================================================================
                     dcc.Tab(label='Gráfico',
                             children=[
                                 dcc.Graph(id='gráfico-datos'),
                                 dcc.Markdown(
                                     """
                                     **Nota:** Si los indicadores que descarga tienen una escala numérica muy
                                     distinta, será difícil apreciar algunos de ellos. En este caso puede ocultar
                                     y mostrar indicadores haciendo clic en la leyenda.
                                     """
                                 ),
                             ],
                             ),
                     # PANEL 3: BUSCAR CÓDIGOS================================================================
                     dcc.Tab(label='Buscar códigos',
                             children=[
                                 dcc.Input(
                                     id='buscar-códigos',
                                     placeholder='Escriba acá los términos que desea buscar...',
                                     value='',
                                     style={'padding': 10, 'width': '50%'}
                                 ),
                                 html.Button('Buscar', id='buscar-códigos-button', n_clicks=0),
                                 dcc.RadioItems(
                                     id='tipo-búsqueda',
                                     options=[
                                         {'label': 'frase exacta', 'value': 'frase'},
                                         {'label': 'todas las palabras', 'value': 'todos'},
                                         {'label': 'alguna de las palabras', 'value': 'algunos'}
                                     ],
                                     value='todos',
                                     labelStyle={'display': 'inline-block'}
                                 ),
                                 dcc.Textarea(id='buscar-mensaje',
                                              value='Ingrese los términos deseados en el cuadro de arriba',
                                              readOnly=True,
                                              style={'width': '100%', 'font-size': '1.5em'},
                                              ),
                                 dash_table.DataTable(
                                     id='lista-códigos',
                                     editable=False,
                                     sort_action='native',
                                     style_cell={'textAlign': 'left'},
                                     row_deletable=False,
                                     filter_action='native',
                                     columns=[
                                         {'name': 'Código', 'id': 'codigo'},
                                         {'name': 'Descripción', 'id': 'descripcion'},
                                         {'name': 'Ruta', 'id': 'DESCRIPCION'},
                                         {'name': 'Unidad', 'id': 'Unidad'},
                                         {'name': 'Medida', 'id': 'Medida'},
                                         {'name': 'Periodicidad', 'id': 'periodo'}
                                     ]
                                 )
                             ],
                             ),
                     # PANEL 4: QUIÉN ES ESTE INDICADOR================================================
                     dcc.Tab(label='¿Quién?',
                             children=[
                                 dcc.Input(
                                     id='quién-es-código',
                                     placeholder='Código...',
                                     value='',
                                     style={'padding': 10, 'width': '10%'}
                                 ),
                                 html.Button('Buscar', id='quién-button', n_clicks=0),
                                 html.Div(
                                     children=[
                                         html.H4('Ubicación de cuenta en el catálogo'),
                                         dcc.Textarea(id='quién',
                                                      value='',
                                                      readOnly=True,
                                                      style={'width': '100%', 'font-size': '1.25em', 'height': 300},)
                                     ],
                                     style={'width': '80%'}
                                 ),
                                 html.Div(
                                     children=[
                                         html.H4('Subcuentas en el catálogo'),
                                         dcc.Textarea(id='subcuentas',
                                                      value='',
                                                      readOnly=True,
                                                      style={'width': '100%', 'font-size': '1.25em', 'height': 300},)
                                     ],
                                     style={'width': '80%'}
                                 ),
                             ],
                     ),
                     # PANEL 5: AYUDA======================================================================
                     dcc.Tab(label='Ayuda',
                             children=[
                                 dcc.Markdown(ayuda)
                             ],
                             ), #final tab 'Trend'
                     ]
                 )
        ],
        style={'width': '80%','float': 'right', 'display': 'inline-block'}
    )
    ]
) # fin app layout



@app.callback(
    Output('seleccionar-indicadores', 'data'),
    Input('agregar-codigo-button', 'n_clicks'),
    State('seleccionar-indicadores', 'data'),
    State('seleccionar-indicadores', 'columns'))
def add_row(n_clicks, rows, columns):
    if n_clicks > 0:
        rows.append({c['id']: '' for c in columns})
    return rows


# FUNCIÓN PARA CONSULTAR QUIÉN ES UNA CUENTA
@app.callback(
    Output('quién', 'value'),
    Output('subcuentas', 'value'),
    Input('quién-button', 'n_clicks'),
    State('quién-es-código', 'value'),
)
def quién_subcuentas(n_clicks, código):

    if código in SW.indicadores.index:

        with io.StringIO() as buf, redirect_stdout(buf):
            SW.quien(código)
            quien = buf.getvalue()

        with io.StringIO() as buf, redirect_stdout(buf):
            SW.subcuentas(código)
            subcuentas = buf.getvalue()
    elif código == '':
        quien = subcuentas = f'Escriba el código deseado en el espacio de arriba.'
    else:
        msg = f'El código {código} no aparece en el catálogo de cuentas.'
        msg += '\n\nEsto puede deberse a que: \n\t1. no exista, o \n\t2. el catálogo de este paquete esté desactualizado.'
        quien = subcuentas = msg

    return quien, subcuentas





# FUNCIÓN PARA CONSULTAR EL CATÁLOGO (Buscar códigos)
@app.callback(
    Output('lista-códigos', 'data'),
    Output('buscar-mensaje', 'value'),
    Input('buscar-códigos-button', 'n_clicks'),
    State('buscar-códigos', 'value'),
    State('tipo-búsqueda', 'value'),
)
def mostrar_códigos(n_clicks, frase, tipo):

    if frase:
        # Acomodar la lista de indicadores
        resultados = SW.buscar(**{tipo: frase})

        # Acomodar los datos descargados como un diccionario, para que lo lea dash
        tabla = resultados.reset_index().to_dict(orient='records')
        nrecords = resultados.shape[0]
        msg = f"Hay {nrecords} indicadores que satisface{'n' if nrecords > 1 else ''} su búsqueda." if nrecords \
            else 'No se encontró ningún indicador; intente la búsqueda con otros términos.'

        # dar los resultados
        return tabla, msg
    else:
        msg = "Para buscar indicadores escriba los términos deseados en el espacio proporcionado."
        return [], msg



# FUNCIÓN PARA CONSULTAR LOS DATOS DEL BCCR (Datos y Gráfico)
@app.callback(
    Output('gráfico-datos', 'figure'),
    Output('datos-descargados', 'columns'),
    Output('datos-descargados', 'data'),
    Output('comando', 'value'),
    Input('consultar-datos-button', 'n_clicks'),
    State('seleccionar-indicadores', 'data'),
    State('fecha-primera-observación', 'date'),
    State('fecha-última-observación', 'date'),
    State('frecuencia', 'value'),
    State('func', 'value'),
    State('fillna', 'value'),
)
def display_output(n_clicks, rows, FechaInicio, FechaFinal, freq, func, fillna):


    # Acomodar la lista de indicadores
    indicadores = dict()
    for item in rows:
        indicadores[item['nombre']] = item['código']

    # Interpretar parámetros freq y fillna
    if freq == 'O':
        freq = None

    if fillna == 'no':
        fillna = None

    if func == 'None':
        func = None

    # Descargar los datos del BCCR
    #busqueda_str = "from bccr import SW\n"
    if '' in indicadores.values():  # no consultar al BCCR al arrancar
        datos = pd.DataFrame(columns=indicadores.keys())
        busqueda_str = "SW()"
    else:
        busqueda_str = "SW("
        busqueda_str += ', '.join([f'{key}={val}' for key, val in indicadores.items()]) + ', '


        busqueda_str += f"FechaInicio='{FechaInicio}', " if FechaInicio else ""
        busqueda_str += f"FechaFinal='{FechaFinal}', " if FechaFinal else ""
        busqueda_str += f"func='{func}', " if func else ""
        busqueda_str += f"freq='{freq}', " if freq else ""
        busqueda_str += f"fillna='{fillna}')" if fillna else ")"
        busqueda_str = busqueda_str[:-3] + ")" if busqueda_str[-3:] == ', )' else busqueda_str
        datos = SW(FechaInicio=FechaInicio, FechaFinal=FechaFinal, func=func, freq=freq, fillna=fillna, **indicadores)

    DATOS_PARA_DESCARGAR[0] = datos  # para poder descargarlos después

    # Hacer la figura
    if datos.shape[0]:
        fig = px.line(datos, x=datos.index.to_timestamp(), y=datos.columns, height=800)
    else:
        fig = px.line(x=[0], y=[0])

    # Cambiar índice para mostrarlo bien en la tabla
    datos.index = datos.index.astype(str)

    # Especificar las columnas de la tabla
    columns = [{'name': 'Fecha', 'id': 'fecha' ,'deletable': False, 'renamable': False}]
    columns += [{'name': col, 'id': col ,'deletable': False, 'renamable': False, 'type':'numeric', 'format':Format(precision=4, scheme=Scheme.fixed)} for col in datos]

    # Acomodar los datos descargados como un diccionario, para que lo lea dash
    tabla = datos.reset_index().to_dict(orient='records')

    # dar los resultados
    return fig, columns, tabla, busqueda_str


# FUNCIÓN PARA DESCARGAR LOS DATOS COMO ARCHIVOS A LA COMPUTADORA LOCAL
@app.callback(
    Output("download-dataframe-xlsx", "data"),
    Input("btn-xlsx", "n_clicks"),
    State("file-name", "value"),
    prevent_initial_call=True,
)
def descargar_excel(n_clicks, nombre):
    datos = DATOS_PARA_DESCARGAR[0]
    return send_data_frame(datos.to_excel, f"{nombre}.xlsx", sheet_name="datos")


@app.callback(
    Output("download-dataframe-dta", "data"),
    Input("btn-dta", "n_clicks"),
    State("file-name", "value"),
    prevent_initial_call=True,
)
def descargar_stata(n_clicks, nombre):
    datos = DATOS_PARA_DESCARGAR[0]
    return send_data_frame(datos.to_stata, f"{nombre}.dta")


@app.callback(
    Output("download-dataframe-csv", "data"),
    Input("btn-csv", "n_clicks"),
    State("file-name", "value"),
    prevent_initial_call=True,
)
def descargar_csv(n_clicks, nombre):
    datos = DATOS_PARA_DESCARGAR[0]
    return send_data_frame(datos.to_csv, f"{nombre}.csv")





def GUI(colab=False):
    if colab:
        app.run_server(mode='external')
    else:
        webbrowser.open('http://127.0.0.1:8050/')
        app.run_server(debug=False)


if __name__ == '__main__':
    GUI()


"""
This module contains functions to download data from Banco Central de Costa Rica.  Data is returned as tidy
pandas DataFrames, indexed as time series.


Randall Romero-Aguilar
May 2016
"""

import time
import webbrowser
import pandas as pd

from .utils import findColumnTitles, fixCommas

pd.set_option('display.width', 500)
pd.set_option('display.max_colwidth', 120)


def api(chart, first=None, last=None, excel=True, open=False):
    """
        Builds a valid url to access data from the BCCR website

    Parameters
    ----------
    chart   : A number identifying the BCCR's data table (integer).
    first   : The first year to download (integer, default=None).
    last    : The last year to download (integer, default=None)
    excel   : Whether to export query as Excel file (boolean, default=TRUE)
    open    : Whether to open the table in the computer's browser (boolean, default=FALSE)

    Returns
    -------
        A valid URL to download the data from indicated chart (string).

    Examples
    --------
        1. Get the url to download the consumer price index (chart 9), using default settings

        >>> api(9)
        http://indicadoreseconomicos.bccr.fi.cr/indicadoreseconomicos/Cuadros/frmVerCatCuadro.aspx?CodCuadro=9&Exportar=True&Excel=True

        2. Get the url to download the money supply (M1, chart 125) since 2010

        >>> api(125, 2010)
        http://indicadoreseconomicos.bccr.fi.cr/indicadoreseconomicos/Cuadros/frmVerCatCuadro.aspx?CodCuadro=125&FecInicial=2010/01/01&Exportar=True&Excel=True

        3. Get the url to download the non-tradable CPI (chart 289) between 2010 and 2015

        >>> api(289, 2010, 2015)
        http://indicadoreseconomicos.bccr.fi.cr/indicadoreseconomicos/Cuadros/frmVerCatCuadro.aspx?CodCuadro=289&FecInicial=2010/01/01&FecFinal=2015/12/31&Exportar=True&Excel=True

        4. Get url to download money supply between 2010 and 2015, but in HTML format (as opposed to the default Excel format)

        >>> api(125, 2010, 2015, excel=False)
        http://indicadoreseconomicos.bccr.fi.cr/indicadoreseconomicos/Cuadros/frmVerCatCuadro.aspx?CodCuadro=125&FecInicial=2010/01/01&FecFinal=2015/12/31

        5. Same as before, but using default dates. The open=True option opens your default browser with the selected data.

        >>> api(125, excel=False, open=True)  # opens link in browser
        http://indicadoreseconomicos.bccr.fi.cr/indicadoreseconomicos/Cuadros/frmVerCatCuadro.aspx?CodCuadro=125
    """
    bccr_web = "http://indicadoreseconomicos.bccr.fi.cr/indicadoreseconomicos/"
    bccr_web += "Cuadros/frmVerCatCuadro.aspx?"
    bccr_web += "CodCuadro=%s" % chart
    bccr_web += "&FecInicial=%s/01/01" % first if first else ""
    bccr_web += "&FecFinal=%s/12/31" % last if last else ""
    bccr_web += "&Exportar=True&Excel=True" if excel else ""

    if open:
        webbrowser.open(bccr_web)

    return bccr_web


def downloadChart(chart, first=None, last=None, quiet=False):
    """
        Downloads data from BCCR website.

        This function converts a BCCR html file into a pandas dataframe. Other
        functions take this dataframe and "clean it" to return a tidy dataframe.

    Parameters
    ----------
    chart   : A number identifying the BCCR's data table (integer).
    first   : The first year to download (integer, default=None).
    last    : The last year to download (integer, default=None)
    quiet   : Print download info if False, nothing if True

    Returns
    -------
        data    : A pandas DataFrame with data in the same format as in BCCR website. First column is used to set the
                (row) index and column headers are use to set the columns property
        title   : Chart title (string)
        subtitle : Chart subtitles (string)

    Examples
    --------
        1. Download data on money supply (M1) for 2011 to 2015

        >>> downloadChart(125, 2011, 2015)

        2. Same as before, but without printing download info(name of series, url, retrieved date-time)

        >>> downloadChart(138, 2011, 2015, quiet=True)
    """
    rawdata = pd.read_html(api(chart, first, last), thousands="")[0]
    title, subtitle, subts2 = rawdata.iloc[:3, 0]
    if pd.notnull(subts2):
        subtitle += ' --- %s' % subts2

    rawdata.set_index(0, inplace=True)
    h = findColumnTitles(rawdata)
    rawdata.columns = rawdata.iloc[h]
    rawdata = rawdata.iloc[h+1:].applymap(fixCommas)

    if not quiet:
        info = 'Downloading chart %s:' % chart
        info += ('\n\t' + title) if title else ''
        info += ('\n\t' + subtitle) if subtitle else ''
        info += '\n\tRetrieved %s from:' % time.strftime("%c")
        info += '\n\t' + api(chart, first, last, excel=False) + '\n'
        print(info)

    rawdata._metadata = {'title': title, 'subtitle': subtitle}
    return rawdata


def web(chart):
    """
        Opens the specified chart in the BCCR website, using the default Internet browser.

    Parameters
    ----------
    series   : A number identifying the BCCR's data table (integer).

    Returns
    -------
        None

    Examples
    --------
        1. Open chart 125

        >>> web(125)
    """

    api(chart, first=None, last=None, excel=False, open=True)
    return


""" The bccr package setup.
Based on setuptools

Randall Romero-Aguilar, 2015-2021
"""

from setuptools import setup

with open('README.md', encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='bccr',
    version='2021.05',
    description='Herramientas para descargar datos del Banco Central de Costa Rica',
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='Randall Romero-Aguilar',
    author_email='randall.romero@outlook.com',
    url='http://randall-romero.com/code/bccr',
    license='MIT',
    keywords='BCCR datos',
    packages=['bccr', 'demos', ],
    python_requires='>=3.7',
    install_requires=['pandas', 'numpy', 'anytree',  'requests', 'beautifulsoup4', 'plotly', 'dash', 'dash-extensions', 'jupyter-dash'],
    package_data={'bccr': ['data/indicadores.pkl', 'data/indicators.pkl', 'data/cuadros.pkl']},
    include_package_data=True
)



